/*
	ProWesS reader - a HTML file browser

	PROforma, PROGS Font & Raster Manager
	ProWesS, PROGS Windowing System

	HTML interpretation

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 14 November 1995
	25.11.1995 proper handling of entities
	19.06.1996 support for IMG tag added
	27.09.1997 now "cw" also possible metric for IMG size (cw=column width)
			   pixels are now approximated by pt (this is more predictable
			   on  printer).
			   when reading height, width was overwritten
	11.11.1998 acirc and acute entities where switched, fixed
*/

#include "HTML_h"
#include "HTMLdtd_h"
#include "HyperText_h"

/*	Entity Names
**	------------
**
**	This table must be matched exactly with ALL the translation tables
*/

static const char *entities[HTML_ENTITIES] = {
  "",				/* capital AE diphthong (ligature) */
  "\\Aacute\\", 	/* capital A, acute accent */
  "\\Acircumflex\\",/* capital A, circumflex accent */
  "\\Agrave\\", 	/* capital A, grave accent */
  "",				/* capital A, ring */
  "",				/* capital A, tilde */
  "",				/* capital A, dieresis or umlaut mark */
  "",				/* capital C, cedilla */
  "\\Eth\\",		/* capital Eth, Icelandic */
  "",				/* capital E, acute accent */
  "\\Ecircumflex\\",/* capital E, circumflex accent */
  "\\Egrave\\", 	/* capital E, grave accent */
  "\\Edieresis\\",	/* capital E, dieresis or umlaut mark */
  "\\Iacute\\", 	/* capital I, acute accent */
  "\\Icircumflex\\",/* capital I, circumflex accent */
  "\\Igrave\\", 	/* capital I, grave accent */
  "\\Idieresis\\",	/* capital I, dieresis or umlaut mark */
  "",				/* capital N, tilde */
  "\\Oacute\\", 	/* capital O, acute accent */
  "\\Ocircumflex\\",/* capital O, circumflex accent */
  "\\Ograve\\", 	/* capital O, grave accent */
  "",				/* capital O, slash */
  "",				/* capital O, tilde */
  "",				/* capital O, dieresis or umlaut mark */
  "\\Thorn\\",		/* capital THORN, Icelandic */
  "\\Uacute\\", 	/* capital U, acute accent */
  "\\Ucircumflex\\",/* capital U, circumflex accent */
  "\\Ugrave\\", 	/* capital U, grave accent */
  "",				/* capital U, dieresis or umlaut mark */
  "\\Yacute\\", 	/* capital Y, acute accent */
  "\\aacute\\", 	/* small a, acute accent */
  "",				/* small a, circumflex accent */
  "\\acute\\",
  "",				/* small ae diphthong (ligature) */
  "\\agrave\\", 	/* small a, grave accent */
  "&",				/* ampersand */
  "",				/* small a, ring */
  "",				/* small a, tilde */
  "",				/* small a, dieresis or umlaut mark */
  "\\brokenbar\\",	/* broken vertical bar */
  "",				/* small c, cedilla */
  "\\cedilla\\",
  "\\cent\\",		/* cent sign */
  "",				/* copyright sign (as in Netscape) - PROGS */
  "\\currency\\",	/* currency sign */
  "",
  "",
  "",				/* small e, acute accent */
  "",				/* small e, circumflex accent */
  "",				/* small e, grave accent */
  "\\eth\\",		/* small eth, Icelandic */
  "",				/* small e, dieresis or umlaut mark */
  "\\onehalf\\",
  "\\onequarter\\",
  "\\threequarters\\",
  ">",				/* greater than */
  "",				/* small i, acute accent */
  "",				/* small i, circumflex accent */
  "",				/* inverted exclamation mark */
  "",				/* small i, grave accent */
  "",
  "",				/* small i, dieresis or umlaut mark */
  "\\quoteleft\\",
  "<",				/* less than */
  "\\macron\\",
  "",
  "\\periodcentered\\ ",
  " ",				/* (no break) space */
  "\\logicalnot\\",
  "",				/* small n, tilde */
  "",				/* small o, acute accent */
  "",				/* small o, circumflex accent */
  "",				/* small o, grave accent */
  "\\ordfeminine\\",/* ordinal indicator, feminine */
  "\\ordmasculine\\",/*ordinal indicator, masculine */
  "",				/* small o, slash */
  "",				/* small o, tilde */
  "",				/* small o, dieresis or umlaut mark */
  "\\paragraph\\",
  "\\plusminus\\",
  "`",				/* pound sign */
  "\"",             /* double quote sign  - June 1994 */
  "\\quoteright\\",
  "\\trademark\\",	/* trademark (as in Netscape) - PROGS */
  "\\section\\",	/* section sign */
  "\\hyphen\\",
  "\\onesuperior\\",
  "\\twosuperior\\",
  "\\threesuperior\\",
  "\\germandbls\\", /* small sharp s, German (sz ligature) */
  "\\thorn\\",		/* small thorn, Icelandic */
  "\\multiply\\",
  "",				/* small u, acute accent */
  "",				/* small u, circumflex accent */
  "",				/* small u, grave accent */
  "\\dieresis\\",	/* umlaut or dieresis */
  "",				/* small u, dieresis or umlaut mark */
  "\\yacute\\", 	/* small y, acute accent */
  "\\yen\\",		/* yen sign */
  "\\ydieresis\\",	/* small y, dieresis or umlaut mark */
};


/* whitespace handling */
/* the whitespace after commands is ignored (SPACE mode) except when COPY */
/* when not in copy mode all space sequences are reduced to one space */
/* when in COPY mode, the newlines are translated to line breaks */

typedef enum {
	TITLE, INTITLE, COMMENT, COPY, INCOPY, INWORD, SPACE
	} ParseMode;

typedef struct _HTStructured {
	HTStructuredClass *isa;
	HyperText *text;
	Style *look;
	ParseMode mode;
	char printing;					/* TRUE when printing ! */
	char printsub;
};

Error HTMLFlush(HTStructured *me)
{
	return HTEnd(me->text);
}

Error HTMLFree(HTStructured *me)
{
	HTEnd(me->text);
	return MEMRelease(me);
}

Error HTMLAbort(HTStructured *me, Error err)
{
	HTEnd(me->text);
	return MEMRelease(me);
}

Error HTMLPutc(HTStructured *me, char ch)
{
	Error err=ERR_OK;

	/* have to make sure that the character set is somewhat conforming !! */
	if (ch=='`') ch='';

	switch (me->mode)
	{
	case TITLE:
		if (ch==' ' || ch=='\n' || ch=='\t')
			break;
		me->mode=INTITLE;
	case INTITLE:
		if (ch=='\n' || ch=='\t') ch=' ';
		if (me->text->posintitle==TITLEBUFFER-1)
		{
			me->mode=COMMENT;
			break;
		}
		me->text->title[me->text->posintitle++]=ch;
		break;

	case COMMENT:
		break;

	case COPY:
/*		  if (ch==' ' || ch=='\n' || ch=='\t')
			break;
		me->mode=INCOPY; */
	case INCOPY:
		if (ch=='\n')
		{
			/* add a space to force empty lines to be visible */
			catch( HTAddCharacter(me->text,' ') );
			err=HTLineBreak(me->text);
		}
		else
		{
			if (ch=='\\')
				err=HTAddCharacter(me->text,ch);
			err=HTAddCharacter(me->text,ch);
		}
		break;

	case INWORD:
		if (ch==' ' || ch=='\n' || ch=='\t')
		{
			err=HTAddCharacter(me->text,' ');
			me->mode=SPACE;
		}
		else
		{
			if (ch=='\\')
				err=HTAddCharacter(me->text,ch);
			err=HTAddCharacter(me->text,ch);
		}
		break;

	case SPACE:
		if (ch!=' ' && ch!='\n' && ch!='\t')
		{
			if (ch=='\\')
				err=HTAddCharacter(me->text,ch);
			err=HTAddCharacter(me->text,ch);
			me->mode=INWORD;
		}
		break;
	}
	return err;
}

Error HTMLPuts(HTStructured *me, const char *str)
{
	Error err=ERR_OK;
	while (*str)
		catch( HTMLPutc(me,*str++) );
	return err;
}

Error HTMLPutb(HTStructured *me, const char *str, Size length)
{
	Error err=ERR_OK;
	while (length--)
		catch( HTMLPutc(me,*str++) );
	return err;
}

static Error Title(HyperText *text, style_h *look)
{
	Error err;
	catch( HTStyleNew(text) );
	catch( HTSpaceVertical(text,look->space_above) );
	catch( HTChangeFont(text,look->font,look->fontsize) );
	err=HTChangeMargin(text, look->indentation, 0);
	return err;
}

Error HTMLElementStart(HTStructured *me, int element_number,
			const char *	attribute_present,
			const char **	attribute_value)
{
	Error err=ERR_OK;

	switch(element_number)
	{
	case HTML_BR:
		err=HTLineBreak(me->text);
		me->mode=SPACE;
		break;

	case HTML_COMMENT:
		me->mode=COMMENT;
		break;

	case HTML_HR:
		catch( HTLineHorizontal(me->text) );
		me->mode=SPACE;
		break;

	case HTML_P:
		err=HTSpaceVertical(me->text,me->look->p_distance);
		err=HTSpaceHorizontal(me->text,me->look->p_indentation);
		me->mode=SPACE;
		break;

	case HTML_TITLE:
		me->mode=TITLE;
		break;

	case HTML_H1 :
		err=Title(me->text,&me->look->h1);
		me->mode=SPACE;
		break;
	case HTML_H2 :
		err=Title(me->text,&me->look->h2);
		me->mode=SPACE;
		break;
	case HTML_H3 :
		err=Title(me->text,&me->look->h3);
		me->mode=SPACE;
		break;
	case HTML_H4 :
		err=Title(me->text,&me->look->h4);
		me->mode=SPACE;
		break;
	case HTML_H5 :
		err=Title(me->text,&me->look->h5);
		me->mode=SPACE;
		break;
	case HTML_H6 :
		err=Title(me->text,&me->look->h6);
		me->mode=SPACE;
		break;
	case HTML_H7 :
		err=Title(me->text,&me->look->h7);
		me->mode=SPACE;
		break;

	case HTML_PRE:
	case HTML_LISTING:
	case HTML_XMP:
	case HTML_PLAINTEXT:
		catch( HTStyleNew(me->text) );
		me->text->style->parsemode=me->mode=COPY;
		catch( HTLineWrap(me->text,FALSE) );
		catch( HTLineBreak(me->text) );
		catch( HTChangeFont(me->text,TYPEWRITER,0) );
		break;

	case HTML_BLOCKQUOTE:
	case HTML_BQ:
	case HTML_ADDRESS:
		catch( HTStyleNew(me->text) );
		catch( HTLineBreak(me->text) );
		catch( HTChangeFont(me->text,me->look->bq_font,0) );
		err=HTChangeMargin(me->text,
				me->look->bq_marginleft, me->look->bq_marginright);
		break;

	case HTML_B:
	case HTML_STRONG:
		catch( HTStyleNew(me->text) );
		catch( HTChangeFont(me->text,BOLD,0) );
		break;

	case HTML_I:
	case HTML_EM:
	case HTML_CITE:
		catch( HTStyleNew(me->text) );
		catch( HTChangeFont(me->text,ITALIC,0) );
		break;

	case HTML_TT:
	case HTML_CODE:
	case HTML_SAMP:
	case HTML_KBD:
	case HTML_VAR:
		catch( HTStyleNew(me->text) );
		catch( HTChangeFont(me->text,TYPEWRITER,0) );
		break;

	case HTML_LH:
		catch( HTStyleNew(me->text) );
		catch( HTChangeFont(me->text,me->look->lh_font,me->look->lh_fontsize) );
		me->mode=SPACE;
		break;

	case HTML_DL:
	case HTML_DIR:
	case HTML_MENU:
	case HTML_OL:
	case HTML_UL:
		catch( HTStyleNew(me->text) );
		catch( HTChangeMargin(me->text,me->look->dl_indentation,0) );
		catch( HTSpaceVertical(me->text, me->look->dl_spacing_vertical) );
		me->mode=SPACE;
		me->text->style->level=0;
		me->text->style->context=element_number;
		break;

	case HTML_LI:
		catch( HTSpaceVertical(me->text, me->look->dl_spacing_vertical) );
		if (me->text->style->context==HTML_OL || me->text->style->context==HTML_UL)
		{
			char *str,number[20];

			catch( HTSpaceHorizontal(me->text, -me->look->dl_indentation) );
			if (me->text->style->context==HTML_OL)
			{
				STRFormatPut(number,"%d.",++me->text->style->level);
				str=number;
			}
			else
			{
				str=me->look->bullet;
			}
			while (*str) catch( HTAddCharacter(me->text,*str++) );
			catch( HTChangeMargin(me->text,0,0) );
		}
		me->mode=SPACE;
		break;

	case HTML_DT:
		catch( HTChangeFont(me->text,me->look->dl_term_font,0) );
		catch( HTSpaceVertical(me->text, me->look->dl_spacing_vertical) );
		catch( HTSpaceHorizontal(me->text, -me->look->dl_indentation) );
		me->mode=SPACE;
		break;
	case HTML_DD:
		catch( HTChangeFont(me->text,ROMAN,0) );
		catch( HTSpaceHorizontal(me->text, me->look->dl_spacing_horizontal) );
		me->mode=SPACE;
		break;

	case HTML_A:
		if (attribute_present[HTML_A_HREF])
		{
			char filename[IO_MAXFULLNAME];
			char *pos; const char *pp;

			STRCopySafe(attribute_value[HTML_A_HREF],filename,IO_MAXFULLNAME);
			for (pos=filename ; *pos ; pos++)
			{
				if (*pos=='\\' || *pos=='/' || *pos=='.') *pos='_';
				if (*pos=='#') *pos='\0';
			}
			for (pp=attribute_value[HTML_A_HREF] ; *pp && *pp!='#' ; pp++);
			if (*pp=='#') pp++;

			if (!me->printing)
				err=HTLinkStart(me->text,filename,pp);
			else
			{
				if ( me->printsub && (
					  attribute_present[HTML_A_PRINT] ||
					( attribute_present[HTML_A_REL] &&
					  STRSameCI(attribute_value[HTML_A_REL],"subdocument") ) ||
					( attribute_present[HTML_A_REV] &&
					  STRSameCI(attribute_value[HTML_A_REV],"toC") ) ) )
					err=HTContentsAdd(me->text,filename);
			}
		}
		if (attribute_present[HTML_A_NAME])
		{
			err=HTLinkMarker(me->text,attribute_value[HTML_A_NAME]);
		}
		break;

	case HTML_IMG:
		{
			pt width=0, height=0; double db;
			if (attribute_present[HTML_IMG_WIDTH])
			{
				STRFormatGet(attribute_value[HTML_IMG_WIDTH],NULL,"%f",&db);
				width=double2pt(db);
			}
			if (attribute_present[HTML_IMG_HEIGHT])
			{
				STRFormatGet(attribute_value[HTML_IMG_HEIGHT],NULL,"%f",&db);
				height=double2pt(db);
			}
			if (attribute_present[HTML_IMG_UNITS])
			{
				if (STRSameCI(attribute_value[HTML_IMG_UNITS],"en"))
				{
					pt size=me->text->lines.prev->styleline->size/2;
					width=fixmul(width,size);
					height=fixmul(height,size);
				}
				else
				if (STRSameCI(attribute_value[HTML_IMG_UNITS],"cw"))
				{
					pt size=me->text->style->marginright-me->text->style->marginleft;
					width=fixmul(width,size);
					height=fixmul(height,size);
				}
			}
			if (attribute_present[HTML_IMG_SRC])
			{
				char filename[IO_MAXFULLNAME];
				char *pos;

				STRCopySafe(attribute_value[HTML_IMG_SRC],filename,IO_MAXFULLNAME);
				for (pos=filename ; *pos ; pos++)
					if (*pos=='\\' || *pos=='/' || *pos=='.') *pos='_';
				err=HTImage(me->text,filename,width,height);
			}
/*			  if (attribute_present[HTML_IMG_ALT]) */
/*			  {                                    */
/*			  } 								   */
		}
		break;
	}

	return err;
}

Error HTMLElementEnd(HTStructured *me, int element_number)
{
	Error err=ERR_OK;

	switch(element_number)
	{
	case HTML_COMMENT:
	case HTML_TITLE:
		me->mode=(ParseMode)me->text->style->parsemode;
		break;

	case HTML_H1:
		catch( HTStyleRestore(me->text) );
		catch( HTSpaceVertical(me->text,me->look->h1.space_below) );
		me->mode=(ParseMode)me->text->style->parsemode;
		break;
	case HTML_H2:
		catch( HTStyleRestore(me->text) );
		catch( HTSpaceVertical(me->text,me->look->h2.space_below) );
		me->mode=(ParseMode)me->text->style->parsemode;
		break;
	case HTML_H3:
		catch( HTStyleRestore(me->text) );
		catch( HTSpaceVertical(me->text,me->look->h3.space_below) );
		me->mode=(ParseMode)me->text->style->parsemode;
		break;
	case HTML_H4:
		catch( HTStyleRestore(me->text) );
		catch( HTSpaceVertical(me->text,me->look->h4.space_below) );
		me->mode=(ParseMode)me->text->style->parsemode;
		break;
	case HTML_H5:
		catch( HTStyleRestore(me->text) );
		catch( HTSpaceVertical(me->text,me->look->h5.space_below) );
		me->mode=(ParseMode)me->text->style->parsemode;
		break;
	case HTML_H6:
		catch( HTStyleRestore(me->text) );
		catch( HTSpaceVertical(me->text,me->look->h6.space_below) );
		me->mode=(ParseMode)me->text->style->parsemode;
		break;
	case HTML_H7:
		catch( HTStyleRestore(me->text) );
		catch( HTSpaceVertical(me->text,me->look->h7.space_below) );
		me->mode=(ParseMode)me->text->style->parsemode;
		break;

	case HTML_PRE:
	case HTML_LISTING:
	case HTML_XMP:
	case HTML_PLAINTEXT:
	case HTML_BLOCKQUOTE:
	case HTML_BQ:
	case HTML_ADDRESS:
	case HTML_LH:
		catch( HTStyleRestore(me->text) );
		catch( HTLineBreak(me->text) );
		me->mode=(ParseMode)me->text->style->parsemode;
		break;

	case HTML_B:
	case HTML_STRONG:
	case HTML_I:
	case HTML_EM:
	case HTML_CITE:
	case HTML_TT:
	case HTML_CODE:
	case HTML_SAMP:
	case HTML_KBD:
	case HTML_VAR:
		catch( HTStyleRestore(me->text) );
		break;

	case HTML_DL:
	case HTML_DIR:
	case HTML_MENU:
	case HTML_OL:
	case HTML_UL:
		catch( HTStyleRestore(me->text) );
		catch( HTSpaceVertical(me->text, me->look->dl_spacing_vertical) );
		me->mode=(ParseMode)me->text->style->parsemode;
		break;

	case HTML_A:
		HTLinkEnd(me->text);
		break;

	/* ..... */
	}

	return err;
}

Error HTMLEntityPut(HTStructured *me, int entity_number)
{
	Error err;
	char *entity=entities[entity_number];

	/* translate the entity number and add to text */
	while (*entity)
		catch( HTAddCharacter(me->text,*entity++) );
	return err;
}

HTStructuredClass HTMLClass={
	"HTML Parser",
	HTMLFlush,
	HTMLFree,
	HTMLAbort,
	HTMLPutc,
	HTMLPuts,
	HTMLPutb,
	HTMLElementStart,
	HTMLElementEnd,
	HTMLEntityPut
};

Error HTML_new(HyperText *text, char printing, char printsub,
			   HTStructured **ret)
{
	Error err;
	HTStructured *context;

	catch( MEMAllocate(sizeof(HTStructured),&context) );

	context->isa=&HTMLClass;
	context->text=text;
	context->printing=printing;
	context->printsub=printsub;
	context->look=text->look;
	text->style->parsemode=SPACE;
	context->mode=SPACE;
	*ret=context;

	return err;
}
